import java.awt.*;
import java.awt.event.*;
import java.applet.*;

public final class
Bounce
    extends Applet
    implements WindowListener
{
    // Frame rate we try to achieve:
    public final static long kFPS = 50;
    public final static int kBallCount = 25;

    // Should naturally synchronize (access to) these but:
    private BounceCanvas mCanvas;
    private boolean mRunning = false;
    private BounceUpdater mUpdater = null;

    public static void main(String [] args)
    {
	Frame f = new Frame("Bounce");
	Bounce b = new Bounce(f);

	b.init();
	b.start();
    }

    public void init()
    {
    }

    public void start()
    {
	if (mUpdater != null) {
	    mUpdater.setRunStatus(false);
	}
	//mCanvas.paint(mCanvas.getGraphics());
	mCanvas.repaint();
	mUpdater = new BounceUpdater(kFPS, mCanvas);
	mUpdater.setRunStatus(true);
	new Thread(mUpdater).start();
    }

    public void stop()
    {
	if (mUpdater != null) {
	    mUpdater.setRunStatus(false);
	    mUpdater = null;
	}
	mCanvas.cleanUp();
    }

    public Bounce()
    {
	this(null);
    }

    public Bounce(Frame f)
    {
	mCanvas = new BounceCanvas(kBallCount);
	setLayout(new BorderLayout());
	add(mCanvas,"Center");

	// Applet?
	if (f == null) {
	} else {
	    f.addWindowListener(this);
	    f.add(this,"Center");
	    f.pack();
	    f.setSize(640,480);
	    f.setLocation(100,100);
	    f.setVisible(true);
	}
    }

    /*** Implementation of WindowListener: ****/
    public void windowClosing(WindowEvent e)
    {
	 System.exit(0);
    }
    public void windowOpened(WindowEvent e) { }
    public void windowClosed(WindowEvent e) { }
    public void windowActivated(WindowEvent e) { }
    public void windowDeactivated(WindowEvent e) { }
    public void windowIconified(WindowEvent e) { }
    public void windowDeiconified(WindowEvent e) { }
}

final class
BounceUpdater
    implements Runnable
{
    private final long mDelay;
    private final BounceCanvas  mComponent;

    private boolean mRunning;

    public BounceUpdater(long FPS, BounceCanvas c)
    {
	mDelay = 1000L / FPS; // max delay
	mComponent = c;
    }

    public synchronized void setRunStatus(boolean b)
    {
	mRunning = b;
    }

    public void run()
    {
	// When running
	while (mRunning) {
	    long startTime = System.currentTimeMillis();

	long time1 = System.currentTimeMillis();
	    mComponent.paint(mComponent.getGraphics());
	long time2 = System.currentTimeMillis();
//System.err.println(" redraw="+(time2-time1)+" ms");
	mComponent.recalc();

	    // Time taken by drawing:
	    long currDelay = (startTime + mDelay) - System.currentTimeMillis();

	    /* Assuming system's clock resolution is 10 ms (true for linux
	     * on x86), let's not even try smaller delays:
	     */
//System.out.println(currDelay);
	    if (currDelay >= 10) {
		try {
		    Thread.sleep(currDelay);
		} catch (InterruptedException ie) { }
	    } else {
		// Let's still try to be nice citizens...
		Thread.yield();
	    }
	}
    }

}

final class
BounceCanvas
    extends Canvas
{
    private Dimension mWindowSize = null;
    private Image mImage = null;
    private Graphics mGfx = null;
    private boolean mInitialized = false;

    public final int mNrOfBalls;

    final BounceBall [] mBalls;

    public BounceCanvas(int ballCount)
    {
	mBalls = new BounceBall[mNrOfBalls = ballCount];
	mInitialized = false;
    }

    public void doInit(Dimension d)
    {
	for (int i = 0; i < mBalls.length; ++i) {
	    double speed = 5.0 + 25.0 * Math.random();
	    if (Math.random() > 0.5) {
		speed = -speed;
	    }
	    int colourType = (int) (3.0 * Math.random());
	    mBalls[i] = new BounceBall(Color.getHSBColor(
				       /*
				       Color.getHSBColor((float)Math.random(),
				       */
				       //(float)((double)colourType / 3.0),
					(float) (Math.random()),
					(float)(0.5+Math.random()/2.0), 1.0f
				       ),
			   0.1 + 0.8 * (double) i / (double)mBalls.length,
					speed,
				       Math.random() * 0.6 + 0.4,
				       6 + (4 * i / mBalls.length),
				       //6,
				       d);
	}
    }

    public void cleanUp()
    {
	if (mImage != null) {
	    Image tmp = mImage;
	    mImage = null;
	    tmp.flush();
	    if (mGfx != null) {
		mGfx.dispose();
		mGfx = null;
	    }
	}
    }

    public void update(Graphics g) { paint(g); }

    public void
    recalc()
    {
	if (!mInitialized) {
	    return;
	}
	Dimension d = getSize();

	if (d == null) {
	    return;
	}
	int w = d.width;
	int h = d.height;

	long time = System.currentTimeMillis();

	/* First we'll calc (possible) new locations:
	 */
	for (int i = mBalls.length; --i >= 0;) {
	    mBalls[i].recalc(w, h, time);
	}

	long time2 = System.currentTimeMillis();
//System.err.print("recalc="+(time2-time)+" ms");

	/* Then check collisions; need to do half of N^2 search...
	 * Which may mean that a ball collides more than once per round:
	 */
	for (int i = mBalls.length; --i >= 0;) {
	    for (int j = i; --j >= 0;) {
		mBalls[i].checkCollisionWith(mBalls[j]);
	    }
	    mBalls[i].updateValues();
	}

	long time3 = System.currentTimeMillis();	
//System.err.print(" checkColl="+(time3-time2)+" ms");
    }

    public synchronized void
    paint(Graphics g)
    {
	Dimension d = getSize();
	int w = d.width;
	int h = d.height;

	if (w <= 0 || h <= 0) {
	    return;
	}
	if (!mInitialized) {
	    doInit(d);
	    mInitialized = true;
	}

	if (mWindowSize == null) {
	    mWindowSize = d;
	} else {
	    if (mWindowSize.width != w || mWindowSize.height != h) {
		mWindowSize = d;
		mImage.flush();
		mGfx.dispose();
		mImage = null;
		mGfx = null;
	    }
	}

	if (mImage == null) {
	    mImage = createImage(w,h);
	    mGfx = mImage.getGraphics();

	    if (mImage == null || mGfx == null) {
		System.exit(1);
	    }
	}

	mGfx.setColor(Color.black);
	mGfx.fillRect(0, 0, w, h);

	// And the actual redraw is simple:
	for (int i = mBalls.length; --i >= 0;) {
	    mBalls[i].paintMe(mGfx, w, h);
	}

	g.drawImage(mImage, 0, 0, this);
    }
}
