import java.awt.*;

public final class
BounceBall
{
    // "Physical" constants:
    //private final static double kGravity = 50.0; // Gravity, in pixels/sec2
    private final static double kGravity = 70.0; // Gravity, in pixels/sec2

    // Properties initialised during construction:
    private final Color mColour, mDimColour;
    private final Color mBrightColour;
    private final int mRadius; // radius in pixels
    private final double mDRadius;

    // State data:
    private long mLastUpdate; // Last update point, milliseconds

    private double mXSpeed, mYSpeed;
    private double mXCoordinate, mYCoordinate;

    private double mNewX, mNewY;

    public BounceBall(Color colour,double xloc,double horizSpeed,
		      double energy,int radius, Dimension windowSize)
    {
	double width = (double) (windowSize.width - 2 * radius);
	double height = (double) (windowSize.height - 2 * radius);

	mRadius = radius;
	mDRadius = (double)radius;
	mColour = colour;
	//mDimColour = colour.darker().darker();
	mDimColour = colour.darker();
	//mBrightColour = colour.brighter().brighter();
	mBrightColour = Color.white;

	// X-location is passed by caller (0.0 - 1.0, from left to right):
	mXSpeed = horizSpeed;
	if (xloc < 0.0) {
	    xloc = 0.0;
	} else if (xloc > 1.0) {
	    xloc = 1.0;
	}
	mXCoordinate = mNewX = width * xloc;

	/* We can choose any single point we wish between max height (energy
	 * multiplied by window height; y speed 0) and 0.0 (max negative
	 * y speed). Potential energy Ep = mgh, for this case mass can be
	 * skipped, getting ep = height * gravity:
	 */
	double relHeight = Math.random();
	height *= energy; // Now height means peak height at bounce
	mYCoordinate = mNewY = height * relHeight;
	double vEnergy = kGravity * (height - mYCoordinate);
	mYSpeed = -Math.sqrt(2.0 * vEnergy);

	mLastUpdate = System.currentTimeMillis();
    }

    public void recalc(int w, int h, long time)
    {
	long timeDiff = time - mLastUpdate;

	double width = (double)w - mDRadius;
	double height = (double)h - mDRadius;

	mLastUpdate = time;
	/* This should always be relatively small, but if we ever
	 * get laggy (disk swapping heavily etc) let's restrict
	 * jerkiness of movement (ie. combination of slow motion
	 * and jerkiness, not either of). So delays above 1 second
	 * are just truncated to 1 sec.
	 */
	if (timeDiff > 1000L) {
	    timeDiff = 1000L;
	}
	double msecDiff = (double)timeDiff / 1000.0;

	// Vertical bounce?
	mNewX = mXCoordinate + (msecDiff * mXSpeed);
	/* Need to add a small constant (say, 0.1) so that even if the vertical
	** speed is close to zero, a ball can't get 'glued' to the edge:
	 */
	if (mNewX <= mDRadius) {
	    mNewX = mDRadius + 0.01;
	    mXSpeed = -mXSpeed;
	} else if (mNewX >= width) {
	    mNewX = width - 0.01;
	    mXSpeed = -mXSpeed;
	} 

	// Bounce? (ie. hit the ground)
	double oldSpeed = mYSpeed;

	mYSpeed -= (kGravity * msecDiff);
	mNewY = mYCoordinate + msecDiff * (oldSpeed + mYSpeed) / 2.0;

	if (mNewY <= mDRadius) {
	    mNewY = mDRadius + 0.01;
	    mYSpeed = -mYSpeed;
	}
	/* Whether to bounce from ceiling...?
	 */
	else if (mNewY >= height) {
	    mNewY = height - 0.01;
	    mYSpeed = -mYSpeed;
	}
    }

    private double getNewDistWithSquared(BounceBall otherBall)
    {
	double dx = otherBall.mNewX - mNewX;
	double dy = otherBall.mNewY - mNewY;

	return (dx * dx + dy * dy);
    }

    private double getOldDistWithSquared(BounceBall otherBall)
    {
	double dx = otherBall.mXCoordinate - mXCoordinate;
	double dy = otherBall.mYCoordinate - mYCoordinate;

	return (dx * dx + dy * dy);
    }

    private double getOldNewDistSquared(BounceBall otherBall)
    {
	double dx = otherBall.mNewX - mXCoordinate;
	double dy = otherBall.mNewY - mYCoordinate;

	return (dx * dx + dy * dy);
    }

    private boolean aboutSameDirection(BounceBall otherBall)
    {
	/* Using dot product we can see if the angle between speed
	 * vectors is less than 180 degrees. If it is, we are heading
	 * towards about the same direction:
	 */
	if (((mXSpeed * otherBall.mXSpeed) + (mXSpeed * otherBall.mXSpeed))
	    < 0.0) {
	    return false;
	}
	return true;
    }

    /* Method for 'undoing' part of the current move; moves new location
     * closer to old location is indicated by the argument (0.0 move
     * completely back to old, 1.0 don't move at all, values between
     * proportional move)
     */
    private void scaleCurrentMove(double factor)
    {
	double invFactor = 1.0 - factor;

	mNewX = (invFactor * mNewX) + (factor * mXCoordinate);
	mNewY = (invFactor * mNewY) + (factor * mYCoordinate);
    }

    private double getSpeed()
    {
	return Math.sqrt(mXSpeed * mXSpeed + mYSpeed * mYSpeed);
    }

    /* Returns true if balls collide; if so, will also adjust ball
     * velocities (and locations) appropriately
     */
    public boolean checkCollisionWith(BounceBall otherBall)
    {
	double minDistSq = otherBall.mDRadius + mDRadius;

	minDistSq *= minDistSq;
	double currDistSq = getNewDistWithSquared(otherBall);

	// No collision, let's leave:
	if (currDistSq > minDistSq) {
	    return false;
	}

	/* Let's 'move back' the balls so they are exactly next
	 * to each other without overlapping. The method should work
	 * ok, except if balls have 'gone through' each other (may
	 * happen with slow systems, ie. sparse update interval or
	 * with high energy objects). In that case an iterative solution
	 * _might_ work better (but not necessarily perfect either)
	 */

	double oldDist = Math.sqrt(getOldDistWithSquared(otherBall));
	double currDist = Math.sqrt(currDistSq);
	double minDist = Math.sqrt(minDistSq);

	/* Factor of 0.0 means 'use old location', 1.0, 'use new location'
	 * and values between intervening locations (ie. basically factor is
	 * the weight factor of old & new locations)
	 */
	double factor = (minDist - currDist) / (oldDist - currDist);
	scaleCurrentMove(factor);
	otherBall.scaleCurrentMove(factor);

	/* Correction succeeded? Let's see:
	System.err.println("Dist before "+(minDist - currDist)
			   +" after "+(minDist - Math.sqrt(getNewDistWithSquared(otherBall))));
	*/

	/* Ok we have more or less the exact collision coordinates. Then
	 * we can get the collision normal, ie. vector from this ball's
	 * center to the other ball's center. This is used to calculate
	 * part of the kinetic energy transferred to the other ball.
	 * Similar for calculating the transfer to other direction as well;
	 * have to be calculated separately though,. both need to be
	 * calculated before applying either one.
	 */

	// Normalized collision vector (collision normal with length of 1.0):
	double colX = otherBall.mNewX - mNewX;
	double colY = otherBall.mNewY - mNewY;
	// currDist _should_ be minDist; but because of rounding errors etc:
	currDist = Math.sqrt(getNewDistWithSquared(otherBall));
	colX /= currDist;
	colY /= currDist;

	double thisSpeed = getSpeed();
	double otherSpeed = otherBall.getSpeed();
	/* dot product of normalized collision vector & current speed
	 * tell which portion of kinetic energy to transfer (for the
	 * otherBall->this ball, need to negate the collision vector)
	 */
	factor = (mXSpeed * colX + mYSpeed * colY) / thisSpeed;
	double factor2 = (otherBall.mXSpeed * -colX + otherBall.mYSpeed * -colY)
	    / otherSpeed;

	/* Factor will be between -1.0 and 1.0. If it's <= 0.0, no energy
	 * will be transferred... ("bump from behind")
	 */

	/* Add energy to other ball? Note that we are assuming masses of
	 * the balls are equal here, otherwise would need to take mass
	 * differences into account as well
	 */
	double en1 = 0.0, en2 = 0.0;
	{
	    double sp1 = getSpeed(), sp2 = otherBall.getSpeed();
	    //System.err.print("motion="+(int)(sp1+sp2)+" energy="+(int)(sp1*sp1 + sp2*sp2));
	    en1 = sp1*sp1 + sp2*sp2;
	}

	int dir = 0;

	if (factor < 0.0) { factor = factor; } else { dir |= 1; }
	if (factor2 < 0.0) { factor2 = factor2; } else { dir |= 2; }

	if (factor > 0.0 || true) {
	    double delta = factor * thisSpeed * colX;

	    otherBall.mXSpeed += delta;
	    mXSpeed -= delta;
	    delta = factor * thisSpeed * colY;
	    otherBall.mYSpeed += delta;
	    mYSpeed -= delta;
	}

	if (factor2 > 0.0 || true) {
	    double delta = factor2 * otherSpeed * -colX;

	    mXSpeed += delta;
	    otherBall.mXSpeed -= delta;
	    delta = factor2 * otherSpeed * -colY;
	    mYSpeed += delta;
	    otherBall.mYSpeed -= delta;
	}
	{
	    double sp1 = getSpeed(), sp2 = otherBall.getSpeed();
	    en2 = sp1*sp1 + sp2*sp2;
//System.err.println(" this="+(int)sp1+", that="+(int)sp2+" ("+(int)(sp1+sp2)+")");
	    //System.err.print(" motion="+(int)(sp1+sp2)+" energy="+(int)(sp1*sp1 + sp2*sp2));
	    //System.err.println("ediff="+(int)(en2-en1)+" dir="+dir);
	}

	return true;
    }

    public void updateValues()
    {
	mXCoordinate = mNewX;
	mYCoordinate = mNewY;
    }

    public void paintMe(Graphics g, int width, int height)
    {
	int left = (int)mXCoordinate - mRadius;
	int top = (int)(height - mYCoordinate) - mRadius;

	// Original:
	/*
	g.setColor(mColour);
	g.fillOval((int)mXCoordinate - mRadius,
		   (int)(height - mYCoordinate - mDRadius),
		   2 * mRadius, 2 * mRadius);
	g.setColor(mDimColour);
	g.drawOval((int)mXCoordinate - mRadius,
	 	   (int)(height - mYCoordinate - mDRadius),
 		   2 * mRadius, 2 * mRadius);
	*/

	// New:
	g.setColor(mDimColour);
	g.fillOval(left, top, 2 * mRadius, 2 * mRadius);
	g.setColor(mColour);
	g.fillOval(left+2, top+2, mRadius, mRadius);

	g.setColor(mBrightColour);
	
	int offset = mRadius * 2 / 3;

	g.drawLine(left+offset,top+offset,
		   left+offset+1,top+offset
		   );
	g.drawLine(left+offset,top+offset+1,
		   left+offset+1,top+offset+1
		   );

	// Arcs:
	/*
	g.drawArc((int)mXCoordinate - mRadius + 2,
	 	   (int)(height - mYCoordinate - mDRadius) + 2,
	 	   2 * mRadius - 4, 2 * mRadius - 4,
	 	  45, 45);
	*/
    }
}
